/*=============================================================================
	XC_Core script implementation
=============================================================================*/

// Includes.
#include "XC_Core.h"
#include "XC_Commandlets.h"
#include "XC_CoreGlobals.h"

/*-----------------------------------------------------------------------------
	The following must be done once per package (.dll).
-----------------------------------------------------------------------------*/

#define NAMES_ONLY
#define AUTOGENERATE_NAME(name) XC_CORE_API FName XC_CORE_##name;
#define AUTOGENERATE_FUNCTION(cls,idx,name) IMPLEMENT_FUNCTION(cls,idx,name)
#include "XC_CoreClasses.h"
#undef AUTOGENERATE_FUNCTION
#undef AUTOGENERATE_NAME
#undef NAMES_ONLY
void RegisterNames()
{
	static INT Registered=0;
	if(!Registered++)
	{
		#define NAMES_ONLY
		#define AUTOGENERATE_NAME(name) extern XC_CORE_API FName XC_CORE_##name; XC_CORE_##name=FName(TEXT(#name),FNAME_Intrinsic);
		#define AUTOGENERATE_FUNCTION(cls,idx,name)
		#include "XC_CoreClasses.h"
		#undef DECLARE_NAME
		#undef NAMES_ONLY
	}
}

// Package implementation.
IMPLEMENT_PACKAGE(XC_Core);

//Function aliasing

#if _MSC_VER
	#define IMPLEMENT_REDIRECTED_FUNCTION(cls,num,func,othercls) \
		extern "C" DLL_EXPORT Native int##othercls##func = (Native)&cls::func; \
		static BYTE othercls##func##Temp = GRegisterNative( num, int##othercls##func );
#else
	#define IMPLEMENT_REDIRECTED_FUNCTION(cls,num,func,othercls) \
		extern "C" DLL_EXPORT { Native int##othercls##func = (Native)&cls::func; } \
		static BYTE othercls##func##Temp = GRegisterNative( num, int##othercls##func );
#endif

#if _MSC_VER
	#define IMPLEMENT_RENAMED_FUNCTION(cls,num,func,othersymbol) \
		extern "C" DLL_EXPORT Native int##cls##othersymbol = (Native)&cls::func; \
		static BYTE cls##othersymbol##Temp = GRegisterNative( num, int##cls##othersymbol );
#else
	#define IMPLEMENT_RENAMED_FUNCTION(cls,num,func,othersymbol) \
		extern "C" DLL_EXPORT { Native int##cls##othersymbol = (Native)&cls::func; } \
		static BYTE cls##othersymbol##Temp = GRegisterNative( num, int##cls##othersymbol );
#endif

#define ABORT_IF(condition, returnvalue, text) \
	if ( condition ) { debugf( NAME_Warning, text); *(INT*)Result = returnvalue; return; }

/*-----------------------------------------------------------------------------
	UXC_CoreStatics
-----------------------------------------------------------------------------*/

void UXC_CoreStatics::execInvSqrt( FFrame &Stack, RESULT_DECL)
{
	P_GET_FLOAT( C);
	P_FINISH;
	*(FLOAT*)Result = _appInvSqrt( C);
}

void UXC_CoreStatics::execHSize( FFrame &Stack, RESULT_DECL)
{
	P_GET_VECTOR( A);
	P_FINISH;
	*(FLOAT*)Result = appSqrt( A.X*A.X+A.Y*A.Y);
}
IMPLEMENT_RENAMED_FUNCTION(UXC_CoreStatics,-1,execHSize,exechsize);
IMPLEMENT_RENAMED_FUNCTION(UXC_CoreStatics,-1,execHSize,exechSize);

void UXC_CoreStatics::execHNormal( FFrame &Stack, RESULT_DECL)
{
	P_GET_VECTOR( A);
	P_FINISH;
	if ( A.X == 0.0 || A.Y == 0.0 )
		*(FVector*)Result = FVector(0,0,0);
	else
		*(FVector*)Result = _UnsafeNormal2D( A);
}

void UXC_CoreStatics::execUnClock( FFrame &Stack, RESULT_DECL)
{
	FTime Temp = appSeconds();
	P_GET_FLOAT_REF( C);
	P_FINISH;
	*(FLOAT*)Result = Temp - *((FTime*)C);
	*((FTime*)C) = Temp;
}

void UXC_CoreStatics::execClock( FFrame &Stack, RESULT_DECL)
{
	P_GET_FLOAT_REF( C);
	P_FINISH;
	*((FTime*)C) = appSeconds();
}
IMPLEMENT_RENAMED_FUNCTION(UXC_CoreStatics,-1,execClock,execclock);

void UXC_CoreStatics::execOr_ObjectObject(FFrame &Stack, RESULT_DECL)
{
	Stack.Step( Stack.Object, Result);
	P_GET_SKIP_OFFSET(W);
	if ( ! (*(INT*)(Result)) )
	{
		Stack.Step( Stack.Object, Result);
		Stack.Code++;
	}
	else
		Stack.Code += W;

/*	P_GET_ACTOR(A);
	P_GET_SKIP_OFFSET(W);
	if( !A )
	{
		P_GET_ACTOR(B);
		*(AActor**)Result = B;
		Stack.Code++; //DEBUGGER
	}
	else
	{
		*(AActor**)Result = A;
		Stack.Code += W;
	}
*/
}


//Developed for FerBotz
void UXC_CoreStatics::execConnectedDests(FFrame &Stack, RESULT_DECL)
{
	guard(ABotz_NavigBase::execConnectedDests);
	P_GET_OBJECT(ANavigationPoint,Start);
	P_GET_ACTOR_REF(End);
	P_GET_INT_REF(SpecIdx);
	P_GET_INT_REF(Idx);
	P_FINISH;

	INT i = 0;

	PRE_ITERATOR;
		*End = NULL;
		*Idx = -1;
		*SpecIdx = -1;
		if ( !Start ) //Get out of this iterator if no start point
		{
			Stack.Code = &Stack.Node->Script(wEndOffset + 1);
			break;
		}

		while ( i<16 && Start->Paths[i] < 0 ) //Skip invalid paths
			i++;

		if ( i<16 )
		{
			FReachSpec *RS;
			RS = &Start->GetLevel()->ReachSpecs(Start->Paths[i]);
			*End = RS->End;
			*Idx = i;
			*SpecIdx = Start->Paths[i];
			i++;
		}
		else
		{
			Stack.Code = &Stack.Node->Script(wEndOffset + 1);
			break;
		}
	POST_ITERATOR;
	unguard;
}

void UXC_CoreStatics::execAppSeconds( FFrame& Stack, RESULT_DECL )
{
	P_FINISH;
	*(FLOAT*)Result = appSeconds().GetFloat();
}

void UXC_CoreStatics::execGetParentClass( FFrame& Stack, RESULT_DECL )
{
	P_GET_OBJECT(UClass,A);
	P_FINISH;
	*(UClass**)Result = A ? A->GetSuperClass() : NULL;
}

void UXC_CoreStatics::execStringToName( FFrame& Stack, RESULT_DECL )
{
	P_GET_STR( S);
	P_FINISH;
	
	*(FName*)Result = FName( *S);
}


void UXC_CoreStatics::execLocs( FFrame& Stack, RESULT_DECL )
{
	P_GET_STR( A);
	P_FINISH;
	*(FString*)Result = A.Locs();
}

void UXC_CoreStatics::execMakeColor( FFrame& Stack, RESULT_DECL )
{
	Stack.Step( Stack.Object, (BYTE*)Result);
	Stack.Step( Stack.Object, ((BYTE*)Result)+1 );
	Stack.Step( Stack.Object, ((BYTE*)Result)+2 );
	if ( *Stack.Code != EX_EndFunctionParms )
		Stack.Step( Stack.Object, ((BYTE*)Result)+3 );
	P_FINISH;

/*	BYTE CData[4];
	Stack.Step( Stack.Object, &CData);
	Stack.Step( Stack.Object, &CData[1]);
	Stack.Step( Stack.Object, &CData[2]);
	if ( *Stack.Code != EX_EndFunctionParms )
		Stack.Step( Stack.Object, &CData[3]);
	else
		CData[3] = 0;
	P_FINISH;
	*(FColor*)Result = *( (FColor*) ((DWORD) &CData));*/
}

void UXC_CoreStatics::execFindObject( FFrame& Stack, RESULT_DECL )
{
	P_GET_STR(Name);
	P_GET_OBJECT(UClass,Class);
	P_FINISH;

	*(UObject**)Result = StaticFindObject( Class, NULL, *Name );
}

void UXC_CoreStatics::execAllObjects( FFrame& Stack, RESULT_DECL )
{
	// Get the parms.
	P_GET_OBJECT(UClass,objClass);
	P_GET_OBJECT_REF(UObject,obj);
	P_FINISH;

	objClass = objClass ? objClass : UObject::StaticClass();
    TObjectIterator<UObject> It;

	PRE_ITERATOR;
		// Fetch next object in the iteration.
		*obj = NULL;
        while (It && *obj==NULL)
        {
            if (It->IsA(objClass) && !It->IsPendingKill())
            {
                *obj = *It;
            }
            ++It;
        }
		if( *obj == NULL )
		{
			Stack.Code = &Stack.Node->Script(wEndOffset + 1);
			break;
		}
	POST_ITERATOR;
}

IMPLEMENT_CLASS(UXC_CoreStatics);



/*-----------------------------------------------------------------------------
	UBinary functions
-----------------------------------------------------------------------------*/

void UBinarySerializer::execCloseFile(FFrame &Stack, RESULT_DECL)
{
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed CloseFile, no file to close"));
	Archive->Close();
	bWrite = 0;
	Archive = NULL;
	*(UBOOL*)Result = true;
}
IMPLEMENT_RENAMED_FUNCTION(UBinarySerializer,-1,execCloseFile,execcloseFile);

void UBinarySerializer::execOpenFileWrite(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execOpenFileWrite);
	P_GET_STR(SrcFile);
	P_FINISH;

	ABORT_IF( Archive, 0, TEXT("Failed OpenFileWrite, there's already an open file, close it first"));
	ABORT_IF( (*SrcFile)[1] == ':', 0, TEXT("Failed OpenFileWrite, drive letter access is forbidden") );
	INT k = SrcFile.Len();
	INT FailCount = 0;
	for ( INT i=0 ; i<k ; i++ )
	{
		ABORT_IF( (*SrcFile)[i] == ':', 0, TEXT("Failed OpenFileWrite, forbidden character") );
		if ( (*SrcFile)[i] == '.' && (*SrcFile)[i+1] == '.' )
			FailCount++;
		ABORT_IF( FailCount >= 2, 0, TEXT("Failed OpenFileWrite, forbidden directory access") );
	}

	FixFilename(*SrcFile);
	Archive = GXFileManager->CreateFileWriter( *SrcFile, FILEWRITE_EvenIfReadOnly );
	if ( Archive )
	{
		*(UBOOL*)Result = true;
		bWrite = 1;
		return;
	}
	*(UBOOL*)Result = false;
	unguard;
}

void UBinarySerializer::execOpenFileRead(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execOpenFileRead);
	P_GET_STR(SrcFile);
	P_FINISH;

	ABORT_IF( Archive, 0, TEXT("Failed OpenFileRead, there's already an open file, close it first"));
	FixFilename( *SrcFile);
	Archive = GXFileManager->CreateFileReader( *SrcFile );
	*(UBOOL*)Result = (Archive != NULL);
	unguard;
}

void UBinarySerializer::execSerializeInt(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeInt);
	P_GET_INT_REF( I);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeInt, no file"));
	*Archive << *I;
	*(UBOOL*)Result = true;
	unguard;
}

void UBinarySerializer::execSerializeFloat(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeFloat);
	P_GET_FLOAT_REF( F);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeFloat, no file"));
	*Archive << *F;
	*(UBOOL*)Result = true;
	unguard;
}

void UBinarySerializer::execSerializeByte(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeByte);
	P_GET_BYTE_REF( B);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeByte, no file"));
	*Archive << *B;
	*(UBOOL*)Result = true;
	unguard;
}

//Very ugly, override string serializator
static void SerializeStringSafe( FArchive& Ar, FString& SA )
{
	INT ArrayRaw[3] = {0,0,0}; //VC++ 2015 complains | Data[0], Num[1], Max[2]
	TArray<TCHAR>* A = (TArray<TCHAR>*) &ArrayRaw[0];
	appMemcpy( A, &SA, 12);
#ifdef _UNICODE
	INT SaveNum = appIsPureAnsi(*SA) ? A->Num() : -A->Num();
#else
	INT SaveNum = A->Num(); //Linux can only save non-ansi strings
#endif
	Ar << SaveNum;
	if( Ar.IsLoading() )
	{
		if ( ArrayRaw[1] ) //A->Num()
			A->Empty();
		A->Add( Abs(SaveNum) );
		A->Shrink();
		if( SaveNum>=0 )
			for( INT i=0; i<ArrayRaw[1]; i++ )
				{ANSICHAR ACh; Ar << *(BYTE*)&ACh; (*A)(i)=FromAnsi(ACh);}
		else
			for( INT i=0; i<ArrayRaw[1]; i++ )
				{UNICHAR UCh; Ar << UCh; (*A)(i)=FromUnicode(UCh);}
		if( Ar.IsLoading() && ArrayRaw[1]==1 )
			A->Empty();
	}
	else
	{
		if( SaveNum>=0 )
			for( INT i=0; i<ArrayRaw[1]; i++ )
				{ANSICHAR ACh=ToAnsi((*A)(i)); Ar << *(BYTE*)&ACh;}
		else
			for( INT i=0; i<ArrayRaw[1]; i++ )
				{UNICHAR UCh=ToUnicode((*A)(i)); Ar << UCh;}
	}
	appMemcpy( &SA, A, 12);
//	appMemzero( &A, 12); //Not necessary
}

void UBinarySerializer::execSerializeString(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeString);
	P_GET_STR_REF( S);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeString, no file"));
	SerializeStringSafe( *Archive, *S );
	*(UBOOL*)Result = true;
	unguard;
}

void UBinarySerializer::execWriteText(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execWriteText);
	P_GET_STR( S);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed WriteText, no file"));
	ABORT_IF( !Archive->IsSaving(), 0, TEXT("Failed WriteText, file is not in write mode"));

	if ( S.Len() > 0 )
	{
		const ANSICHAR* AC = TCHAR_TO_ANSI( *S );
		Archive->Serialize( (void*)AC, S.Len() );
	}
	*(UBOOL*)Result = true;
	unguard;
}

void UBinarySerializer::execSerializeVector(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeVector);
	P_GET_VECTOR_REF( V);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeVector, no file"));
	*Archive << *V;
	*(UBOOL*)Result = true;
	unguard;
}

void UBinarySerializer::execSerializeRotator(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeRotator);
	P_GET_ROTATOR_REF( R);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeRotator, no file"));
	*Archive << *R;
	*(UBOOL*)Result = true;
	unguard;
}

void UBinarySerializer::execSerializeTo(FFrame &Stack, RESULT_DECL)
{
	guard(UBinarySerializer::execSerializeTo);
	P_GET_OBJECT( UObject, O);
	P_GET_NAME( VariableName);
	P_GET_INT( LimitSize);
	P_FINISH;

	ABORT_IF( !Archive, 0, TEXT("Failed SerializeTo, no file"));
	ABORT_IF( !O, 0, TEXT("Failed SerializeTo, no Object specified"));

	for( TFieldIterator<UProperty> It( O->GetClass() ); It; ++It )
		if ( appStricmp( It->GetName(), *VariableName )== 0 )
		{
			INT aSize = It->ElementSize * It->ArrayDim;
			if ( LimitSize <= 0 || LimitSize > aSize )
				LimitSize = aSize;
			BYTE* Data = (BYTE*)O + It->Offset;
			Archive->Serialize( Data, LimitSize);
			*(UBOOL*)Result = true;
			return;
		}

	debugf( NAME_Warning, TEXT("Failed SerializeTo, Variable %s not found"), VariableName);
	unguard;
}

void UBinarySerializer::execPosition(FFrame &Stack, RESULT_DECL)
{
	P_FINISH;

	ABORT_IF( !Archive, -1, TEXT("Failed Position, no file open") )
	*(INT*)Result = Archive->Tell();
}
IMPLEMENT_RENAMED_FUNCTION(UBinarySerializer,-1,execPosition,execposition);

void UBinarySerializer::execTotalSize(FFrame &Stack, RESULT_DECL)
{
	P_FINISH;

	ABORT_IF( !Archive, -1, TEXT("Failed TotalSize, no file open") )
	*(INT*)Result = Archive->TotalSize();
}
IMPLEMENT_RENAMED_FUNCTION(UBinarySerializer,-1,execTotalSize,exectotalSize);
IMPLEMENT_RENAMED_FUNCTION(UBinarySerializer,-1,execTotalSize,exectotalsize);
IMPLEMENT_RENAMED_FUNCTION(UBinarySerializer,-1,execTotalSize,execTotalsize);


IMPLEMENT_CLASS(UBinarySerializer);

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
