

#include "XC_Core.h"
#include "XC_CoreGlobals.h"
#include "Atomics.h"

#ifdef __LINUX_X86__
#include "FXFileManagerLinux.h"
#endif


XC_CORE_API FMemStack GXCMem;
XC_CORE_API FFileManager* GXFileManager = GFileManager;

//*************************************************
// Pathing in both platforms
// Also sets up the Linux file manager if needed
//*************************************************
XC_CORE_API void FixFilename( const TCHAR* Filename )
{
	TCHAR* Cur;
#ifdef __LINUX_X86__
	if ( GXFileManager == GFileManager )
		GXFileManager = new FXFileManagerLinux;
	for( Cur = (TCHAR*)Filename; *Cur != '\0'; Cur++ )
		if( *Cur == '\\' )
			*Cur = '/';
#else
	for( Cur = (TCHAR*)Filename; *Cur != '\0'; Cur++ )
		if( *Cur == '/' )
			*Cur = '\\';
#endif
}


//*************************************************
// Register GNatives in their respective opcodes
// Helps keeping the engine clean when playing
// on protected GNatives environments
//*************************************************
#pragma DISABLE_OPTIMIZATION
XC_CORE_API void XCCNatives( UBOOL bEnable)
{
	if ( bEnable )
	{
		GetDefault<UXC_CoreStatics>()->bGNatives = true;
		GNatives[198] = (Native)&UXC_CoreStatics::execMakeColor;
		GNatives[238] = (Native)&UXC_CoreStatics::execLocs;
		GNatives[391] = (Native)&UXC_CoreStatics::execStringToName;
		GNatives[600] = (Native)&UXC_CoreStatics::execFindObject;
		GNatives[601] = (Native)&UXC_CoreStatics::execGetParentClass;
		GNatives[602] = (Native)&UXC_CoreStatics::execAllObjects;
		GNatives[643] = (Native)&UXC_CoreStatics::execAppSeconds;
		GNatives[3554] = (Native)&UXC_CoreStatics::execConnectedDests;
		GNatives[3555] = (Native)&UXC_CoreStatics::execOr_ObjectObject;
		GNatives[3556] = (Native)&UXC_CoreStatics::execClock;
		GNatives[3557] = (Native)&UXC_CoreStatics::execUnClock;
		GNatives[3558] = (Native)&UXC_CoreStatics::execGetEnum;
		GNatives[3570] = (Native)&UXC_CoreStatics::execHNormal;
		GNatives[3571] = (Native)&UXC_CoreStatics::execHSize;
		GNatives[3572] = (Native)&UXC_CoreStatics::execInvSqrt;
	}
	else
	{
		GetDefault<UXC_CoreStatics>()->bGNatives = false;
		GNatives[198] = (Native)&UObject::execUndefined;
		GNatives[238] = (Native)&UObject::execUndefined;
		GNatives[391] = (Native)&UObject::execUndefined;
		GNatives[600] = (Native)&UObject::execUndefined;
		GNatives[601] = (Native)&UObject::execUndefined;
		GNatives[602] = (Native)&UObject::execUndefined;
		GNatives[643] = (Native)&UObject::execUndefined;
		GNatives[3554] = (Native)&UObject::execUndefined;
		GNatives[3555] = (Native)&UObject::execUndefined;
		GNatives[3556] = (Native)&UObject::execUndefined;
		GNatives[3557] = (Native)&UObject::execUndefined;
		GNatives[3558] = (Native)&UObject::execUndefined;
		GNatives[3570] = (Native)&UObject::execUndefined;
		GNatives[3571] = (Native)&UObject::execUndefined;
		GNatives[3572] = (Native)&UObject::execUndefined;
	}
}
#pragma ENABLE_OPTIMIZATION


//*************************************************
// Globals initialization
// DeInit must be called always after init
// This helps running the globals in commandlets
// and other objects that aren't controlled by a 
// main engine like XC_Engine
//*************************************************

XC_CORE_API void InitXCGlobals()
{
	guard( XC_Core::InitXCGlobals);
	
	guard( InitMem )
	GXCMem.Init( 65536 );
	unguard;

	unguard;
}

XC_CORE_API void DeInitXCGlobals()
{
	guard( XC_Core::DeInitXCGlobals);
	
	guard( DeInitMem )
	GXCMem.Exit();
	unguard;

	unguard;
}


//*************************************************
// Script stuff
// Finds a (non-state) function in a struct
// Finds a variable in a struct, increments *Found
//*************************************************
XC_CORE_API UFunction* FindBaseFunction( UStruct* InStruct, const TCHAR* FuncName)
{
	FName NAME_FuncName = FName( FuncName, FNAME_Find);
	if ( NAME_FuncName != NAME_None )
	{
		for( TFieldIterator<UFunction> Func( InStruct ); Func; ++Func )
			if( Func->GetFName() == NAME_FuncName )
				return (UFunction*) *Func;
	}
	return NULL;	
}

XC_CORE_API UProperty* FindScriptVariable( UStruct* InStruct, const TCHAR* PropName, INT* Found)
{
	FName NAME_PropName = FName( PropName, FNAME_Find);
	if ( NAME_PropName != NAME_None )
	{
		for( TFieldIterator<UProperty> Prop( InStruct ); Prop; ++Prop )
			if( Prop->GetFName() == NAME_PropName )
			{
				(*Found)++;
				return (UProperty*) *Prop;
			}
	}
	return NULL;	
}


//*************************************************
// String table sorting
//*************************************************
//Caching OList generates 3 more instructions at start
//But also removes two instructions on the appStricmp call

//Dynamic array
XC_CORE_API void SortStringsA( TArray<FString>* List)
{
	SortStringsSA( (FString*) List->GetData(), List->Num() );
/*	FString* OList = (FString*) List->GetData(); 
	for ( INT i=1 ; i<List->Num() ; i++ )
	{
		if ( appStricmp( *OList[i], *OList[i-1]) < 0 )
		{
			//Circular iteration to atomically swap data (no FArray reallocs)
			//Need to port this to assembly!!!!
			INT iStore;
			INT* Ptr = (INT*) &OList[i];
			for ( INT j=0 ; j<3 ; j++ )
			{
				iStore = *Ptr;
				*Ptr = *(Ptr-3);
				*(Ptr-3) = iStore;
				Ptr++;
			}
			if ( i == 1 )
				continue;
			i -= 2;

		}
	}
*/
}

//Static array, optimized for speed
XC_CORE_API void SortStringsSA( FString* List, INT ArrayMax)
{
	for ( INT i=1 ; i<ArrayMax ; i++ )
	{
		if ( appStricmp( *List[i], *List[i-1]) < 0 )
		{
			//Atomically swap data (no FArray reallocs)
			//Compiler does a nice job here
			INT* Ptr = (INT*) &List[i];
			INT iStore = *Ptr;
			*Ptr = *(Ptr-3);
			*(Ptr-3) = iStore;
			iStore = *(Ptr+1);
			*(Ptr+1) = *(Ptr-2);
			*(Ptr-2) = iStore;
			iStore = *(Ptr+2);
			*(Ptr+2) = *(Ptr-1);
			*(Ptr-1) = iStore;
			if ( i == 1 )
				continue;
			i -= 2;
		}
	}
}




//*************************************************
// Multi-threaded log
// Worker thread stores messages in memory
// Main thread flushes messages into log
//*************************************************

#define THREADED_LOG_SPIN_MAX 500
struct FThreadSafeLogEntry
{
	volatile FThreadSafeLogEntry* Next;
	EName NameIndex;
	TCHAR RawLog[800];

	FThreadSafeLogEntry( EName InName, const TCHAR* InStrLog);
	~FThreadSafeLogEntry() {};
};

static volatile FThreadSafeLogEntry* ThreadedLogs = NULL;
static volatile FThreadSafeLogEntry* ThreadedLogsLast = NULL;
static volatile UBOOL bHandlingEntries = 0; //This is our lock

XC_CORE_API void ThreadedLog( EName InName, const TCHAR* InStrLog)
{
	if ( !InStrLog[0] )
		return;
	new FThreadSafeLogEntry( InName, InStrLog);
}

XC_CORE_API void ThreadedLogFlush()
{
	guard( ThreadedLogFlush );
	INT SpinCount = THREADED_LOG_SPIN_MAX * 2;

	UBOOL bLockAcquired = 0;
	while ( !bLockAcquired )
	{
		if ( FPlatformAtomics::InterlockedCompareExchange( &bHandlingEntries, 1, 0) == 0 )
			bLockAcquired = 1; 		//Nobody's handling entries, let's flush
		else
		{
			if ( SpinCount % 20 == 1 )	appSleep( 0.f ); //Critical
			if ( SpinCount-- <= 0 ) //Don't flush
			{
				FPlatformAtomics::InterlockedExchange( &bHandlingEntries, 0); //Dangerous
				return;
			}
		}
	}

	FThreadSafeLogEntry* Link = (FThreadSafeLogEntry*) FPlatformAtomics::InterlockedCompareExchange( (volatile INT*)&ThreadedLogs, 0, 0);
	while ( Link )
	{
		debugf( Link->NameIndex, (const TCHAR*) Link->RawLog);	
		FThreadSafeLogEntry* Temp = Link;
		Link = (FThreadSafeLogEntry*) Link->Next;
		delete Temp;
	}

	FPlatformAtomics::InterlockedExchange( &bHandlingEntries, 0); //Release the lock
	FPlatformAtomics::InterlockedExchange( (volatile INT*)&ThreadedLogs, 0);
	FPlatformAtomics::InterlockedExchange( (volatile INT*)&ThreadedLogsLast, 0);
	unguard;
}

FThreadSafeLogEntry::FThreadSafeLogEntry( EName InName, const TCHAR* InStrLog)
	: Next(NULL) , NameIndex(InName)
{
	appStrncpy( RawLog, InStrLog, 799 );

	//Attach to global threaded log
	volatile INT SpinCount = THREADED_LOG_SPIN_MAX;
	while ( true )
	{
		//Nobody's handling entries, insert ourselves in the chain
		if ( FPlatformAtomics::InterlockedCompareExchange( &bHandlingEntries, 1, 0) == 0 )
		{
			// If we're not first in line, add to last in line
			if ( FPlatformAtomics::InterlockedCompareExchange( (volatile INT*) &ThreadedLogs, (INT)this, 0) != 0 )
				FPlatformAtomics::InterlockedExchangePtr( (volatile void**) &(ThreadedLogsLast->Next), this);
			FPlatformAtomics::InterlockedExchangePtr( (volatile void**) &ThreadedLogsLast, this);
			FPlatformAtomics::InterlockedExchange( &bHandlingEntries, 0); //Release the lock
			return;
		}
		if ( SpinCount == 1 )
			appSleep( 0.f ); //Critical, last resort
		if ( SpinCount-- <= 0 )
		{
			delete this;
			return;
		}
	}

}